--[[-----------------------------------------------------------------------------
Button Widget
Graphical Button.
-------------------------------------------------------------------------------]]
local Healcsi      = Healcsi
local Gui          = Healcsi.Gui
local TopParent    = Healcsi.UIParent
local HiddenParent = Healcsi.HiddenParent

local Type, Version = "Healcsi-Button", 1
if not Gui or (Gui:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local pairs = pairs

-- WoW APIs
local _G = _G
local CreateFrame = CreateFrame

--[[-----------------------------------------------------------------------------
Scripts
-------------------------------------------------------------------------------]]
local function Button_OnClick(frame, ...)
	Gui:ClearFocus()
	frame.obj:Fire("OnClick", ...)
end

local function Button_OnMouseDown(frame, ...)
	frame.hiddenState = frame.state
	frame.state = "pressed"
	frame.obj:ApplyTexture()
end

local function Button_OnMouseUp(frame, ...)
	if (frame.hiddenState) then
		frame.state = frame.hiddenState
		frame.hiddenState = nil
	else
		frame.state = "normal"
	end
	frame.obj:ApplyTexture()
end

local function Control_OnEnter(frame)
	frame.obj:Fire("OnEnter")
	if (frame.state ~= "pressed") then
		frame.state = "hover"
	else
		frame.hiddenState = "hover"
	end
	frame.obj:ApplyTexture()
end

local function Control_OnLeave(frame)
	frame.obj:Fire("OnLeave")
	if (frame.state ~= "pressed") then
		frame.state = "normal"
	else
		frame.hiddenState = "normal"
	end
	frame.obj:ApplyTexture()
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
	["OnAcquire"] = function(self)
		-- restore default values
		self:SetHeight(24)
		self:SetWidth(200)
		self:SetDisabled(false)
		self:SetAutoWidth(false)

		self:SetFont(Healcsi:GetFont("Friz Quadrata TT")) 
		self:SetFontColor(Healcsi:GetColor("White", 0.9))

		self:SetNormalTexture("LiteStep")
		self:SetHoverTexture("LiteStep")
		self:SetPressedTexture("LiteStep")

		self:SetNormalTextureColor(Healcsi:GetColor("DarkBlue", 0.8))
		self:SetHoverTextureColor(Healcsi:GetColor("BluePressed", 0.8))
		self:SetPressedTextureColor(Healcsi:GetColor("BlueSelected", 0.8))

		self:SetText()
	end,

	-- ["OnRelease"] = nil,

	["SetNormalTexture"] = function(self, texture)
		self.normalTexture = Healcsi:GetMedia("statusbar", texture)
		self:ApplyTexture()
	end,

	["SetNormalTextureColor"] = function(self, r, g, b, a)
		self.normalTextureColor = self.normalTextureColor or {}
		self.normalTextureColor.r = r
		self.normalTextureColor.g = g
		self.normalTextureColor.b = b
		self.normalTextureColor.a = a
		self:ApplyTexture()
	end,

	["SetHoverTexture"] = function(self, texture)
		self.hoverTexture = Healcsi:GetMedia("statusbar", texture)
		self:ApplyTexture()
	end,

	["SetHoverTextureColor"] = function(self, r, g, b, a)
		self.hoverTextureColor = self.hoverTextureColor or {}
		self.hoverTextureColor.r = r
		self.hoverTextureColor.g = g
		self.hoverTextureColor.b = b
		self.hoverTextureColor.a = a
		self:ApplyTexture()
	end,

	["SetPressedTexture"] = function(self, texture)
		self.pressedTexture = Healcsi:GetMedia("statusbar", texture)
		self:ApplyTexture()
	end,

	["SetPressedTextureColor"] = function(self, r, g, b, a)
		self.pressedTextureColor = self.pressedTextureColor or {}
		self.pressedTextureColor.r = r
		self.pressedTextureColor.g = g
		self.pressedTextureColor.b = b
		self.pressedTextureColor.a = a
		self:ApplyTexture()
	end,

	["ApplyTexture"] = function(self)
		if (self.frame.state == "normal") then
			self.frame.background:SetTexture(self.normalTexture)
			local color = self.normalTextureColor or {0, 0, 0}
			self.frame.background:SetVertexColor(color.r, color.g, color.b, color.a or 0.9)
		elseif (self.frame.state == "hover") then
			self.frame.background:SetTexture(self.hoverTexture)
			local color = self.hoverTextureColor or {0, 0, 0}
			self.frame.background:SetVertexColor(color.r, color.g, color.b, color.a or 0.9)
		elseif (self.frame.state == "pressed") then
			self.frame.background:SetTexture(self.pressedTexture)
			local color = self.pressedTextureColor or {0, 0, 0}
			self.frame.background:SetVertexColor(color.r, color.g, color.b, color.a or 0.9)
		end
	end,

	["SetFont"] = function(self, font, height, flags)
		self.text:SetFont(font, height or 13, flags)
	end,

	["SetFontColor"] = function(self, r, g, b, a)
		self.text:SetVertexColor(r or 0.1, g or 0.1, b or 0.1, a or 1)

		--self.text:SetShadowOffset(0,1)
		--self.text:SetShadowColor(0.8, 0.8, 0.8, 0.2)
	end,

	["SetText"] = function(self, text)
		self.text:SetText(text)
		if self.autoWidth then
			self:SetWidth(self.text:GetStringWidth() + 30)
		end
	end,
	
	["SetAutoWidth"] = function(self, autoWidth)
		self.autoWidth = autoWidth
		if self.autoWidth then
			self:SetWidth(self.text:GetStringWidth() + 30)
		end
	end,

	["SetDisabled"] = function(self, disabled)
		self.disabled = disabled
		if disabled then
			self:SetFontColor(Healcsi:GetColor("Gray", 0.9))
			self:SetNormalTextureColor(Healcsi:GetColor("Gray", 0.8))
			self:SetHoverTextureColor(Healcsi:GetColor("Gray", 0.8))
			self:SetPressedTextureColor(Healcsi:GetColor("Gray", 0.8))
			self.frame:Disable()
		else
			self:SetFontColor(Healcsi:GetColor("White", 0.9))
			self:SetNormalTextureColor(Healcsi:GetColor("DarkBlue", 0.8))
			self:SetHoverTextureColor(Healcsi:GetColor("BluePressed", 0.8))
			self:SetPressedTextureColor(Healcsi:GetColor("BlueSelected", 0.8))
			self.frame:Enable()
		end
	end
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Constructor()
	local name = "HealcsiGui-Button" .. Gui:GetNextWidgetNum(Type)
	local frame = CreateFrame("Button", name, TopParent)
	frame:Hide()

	frame:EnableMouse(true)
	frame:SetScript("OnClick", Button_OnClick)
	frame:SetScript("OnMouseDown", Button_OnMouseDown)
	frame:SetScript("OnMouseUp", Button_OnMouseUp)
	frame:SetScript("OnEnter", Control_OnEnter)
	frame:SetScript("OnLeave", Control_OnLeave)
	frame.state = "normal"

	local background = frame:CreateTexture(nil, "BACKGROUND")
	frame.background = background
	background:SetAllPoints(frame)
	background:SetTexture(Healcsi:GetMedia("statusbar", "Solid"))
	background:SetVertexColor(0, 0, 0, 1)

	local text = frame:CreateFontString(nil, "ARTWORK")
	text:ClearAllPoints()
	text:SetPoint("TOPLEFT", 15, -1)
	text:SetPoint("BOTTOMRIGHT", -15, 1)
	text:SetJustifyV("MIDDLE")

	local widget = {
		text  = text,
		frame = frame,
		type  = Type
	}
	for method, func in pairs(methods) do
		widget[method] = func
	end

	return Gui:RegisterAsWidget(widget)
end

Gui:RegisterWidgetType(Type, Constructor, Version)
