--[[-----------------------------------------------------------------------------
ColorPicker Widget
-------------------------------------------------------------------------------]]
local Healcsi = Healcsi
local L       = Healcsi.Locale

local Type, Version = "Healcsi-ColorPicker", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local pairs = pairs
local sqrt  = math.sqrt
local sin   = math.sin
local cos   = math.cos
local atan  = math.atan
local abs   = math.abs
local pi    = math.pi

-- WoW APIs
local CreateFrame = CreateFrame

-- Global vars/functions that we don't upvalue since they might get hooked, or upgraded
-- List them here for Mikk's FindGlobals script
-- GLOBALS: ShowUIPanel, HideUIPanel, ColorPickerFrame, OpacitySliderFrame

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]

local function RGBAToHex(r, g, b, a)
	return string.format('%02X%02X%02X%02X', a*255, r*255, g*255, b*255)
end

local function RGBToHex(r, g, b)
	return string.format('%02X%02X%02X', r*255, g*255, b*255)
end

local function HexToRGBA(hex)
	return tonumber('0x' .. string.sub(hex, 3, 4), 10) / 255,
		tonumber('0x' .. string.sub(hex, 5, 6), 10) / 255,
		tonumber('0x' .. string.sub(hex, 7, 8), 10) / 255,
		tonumber('0x' .. string.sub(hex, 1, 2), 10) / 255
end

local function HexToRGB(hex)
	return tonumber('0x' .. string.sub(hex, 1, 2), 10) / 255,
		tonumber('0x' .. string.sub(hex, 3, 4), 10) / 255,
		tonumber('0x' .. string.sub(hex, 5, 6), 10) / 255
end

function round(num, numDecimalPlaces)
	return tonumber(string.format("%." .. (numDecimalPlaces or 0) .. "f", num))
end

local function RadToDeg(rad)
	return rad * (180 / pi)
end

local function DegToRad(degree)
	return degree * (pi / 180)
end

local function IsInsideColorWheel(valueX, valueY)
	local radius = 128
	local x = abs(valueX)
	local y = abs(valueY)

	if ( (x^2) + (y^2) <= (radius^2) ) then
		return true
	end
	return false
end

local function LastInsideCoordinates(valueX, valueY)
	local radius = 128

	local y = sqrt((radius^2) / ( ((valueX^2)/(valueY^2)) + 1 ))
	local x = (y * valueX) / valueY

	if (valueX < 0 and x > 0) then
		x = x * -1
	elseif (valueX > 0 and x < 0) then
		x = x * -1
	end

	if (valueY < 0 and y > 0) then
		y = y * -1
	elseif (valueY > 0 and y < 0) then
		y = y * -1
	end

	return x, y
end

local function GetWheelClickCoordinates()
	if not(Healcsi.ColorPicker) then return end
	
	local screenWidth  = GetScreenWidth()
	local screenHeight = GetScreenHeight()
	local window       = Healcsi.ColorPicker
	local windowW      = window:GetWidth()
	local windowH      = window:GetHeight()

	local newX, newY = 0, 0
	local _, _, relativePoint, windowX, windowY = window:GetPoint()

	if relativePoint == "CENTER" then
		newX = (screenWidth/2) + windowX - (windowW/2)
		newY = (screenHeight/2) + windowY - (windowH/2)
	elseif relativePoint == "BOTTOM" then
		newX = (screenWidth/2) + windowX - (windowW/2)
		newY = windowY
	elseif relativePoint == "TOP" then
		newX = (screenWidth/2) + windowX - (windowW/2)
		newY = screenHeight + windowY - windowH
	elseif relativePoint == "LEFT" then
		newX = windowX
		newY = (screenHeight/2) + windowY - (windowH/2)
	elseif relativePoint == "RIGHT" then
		newX = screenWidth + windowX - windowW
		newY = (screenHeight/2) + windowY - (windowH/2)
	elseif relativePoint == "BOTTOMLEFT" then
		newX = windowX
		newY = windowY
	elseif relativePoint == "BOTTOMRIGHT" then
		newX = screenWidth + windowX - windowW
		newY = windowY
	elseif relativePoint == "TOPLEFT" then
		newX = windowX
		newY = screenHeight + windowY - windowH
	elseif relativePoint == "TOPRIGHT" then
		newX = screenWidth + windowX - windowW
		newY = screenHeight + windowY - windowH
	end

	local scale, cursorX, cursorY = UIParent:GetEffectiveScale(), GetCursorPosition()

	local valueX = (cursorX/scale) - (newX + 19 + 128) -- 19 is left margin, 128 is half of the wheel's width
	local valueY = (cursorY/scale) - (newY + 35 + 128) -- 35 is bottom margin, 128 is half of the wheel's height

	if not(IsInsideColorWheel(valueX, valueY)) then
		valueX, valueY = LastInsideCoordinates(valueX, valueY)
	end

	return valueX, valueY
end

local function GetLightnessClickCoordinate()
	if not(Healcsi.ColorPicker) then return end
	
	local screenWidth  = GetScreenWidth()
	local screenHeight = GetScreenHeight()
	local window       = Healcsi.ColorPicker
	local windowH      = window:GetHeight()

	local newY = 0
	local _, _, relativePoint, _, windowY = window:GetPoint()

	if (relativePoint == "CENTER" or relativePoint == "LEFT" or relativePoint == "RIGHT") then
		newY = (screenHeight/2) + windowY - (windowH/2)
	elseif (relativePoint == "BOTTOM" or relativePoint == "BOTTOMLEFT" or relativePoint == "BOTTOMRIGHT") then
		newY = windowY
	elseif (relativePoint == "TOP" or relativePoint == "TOPLEFT" or relativePoint == "TOPRIGHT")then
		newY = screenHeight + windowY - windowH
	end

	local scale, cursorX, cursorY = UIParent:GetEffectiveScale(), GetCursorPosition()

	local valueY = (cursorY/scale) - (newY + 35 + 256 - 50) -- 35 is bottom margin, 128 is half of the wheel's height

	if (valueY > 0) then valueY = 0 end
	if (valueY < -200) then valueY = -200 end

	return abs(valueY)
end

-- @param h        hue (0-360)
-- @param s        saturation (0.0-1.0)
-- @param V        value (0.0-1.0)
-- @return         an R, G, and B component of RGB
-- http://www.rapidtables.com/convert/color/hsv-to-rgb.htm
local function HSV2RGB(h, s, v)
	local c = v * s
	local x = c * (1 - abs( ((h/60)%2) - 1 ))
	local m = v - c

	local r, g, b

	if (h < 60 or h == 360) then -- 360 would be missing otherwise... i think
		r, g, b = c, x, 0
	elseif (h < 120) then
		r, g, b = x, c, 0
	elseif (h < 180) then
		r, g, b = 0, c, x
	elseif (h < 240) then
		r, g, b = 0, x, c
	elseif (h < 300) then
		r, g, b = x, 0, c
	elseif (h < 360) then
		r, g, b = c, 0, x
	end

	return r+m, g+m, b+m
end

-- http://www.rapidtables.com/convert/color/rgb-to-hsv.htm
local function RGB2HSV(r, g, b)
	local cmax = math.max(r, g, b)
	local cmin = math.min(r, g, b)
	local d = cmax - cmin

	local hue
	if d ==0 then
		hue = 0
	elseif cmax == r then
		hue = 60 * (((g-b)/d)%6)
	elseif cmax == g then
		hue = 60 * (((b-r)/d)+2)
	elseif cmax == b then
		hue = 60 * (((r-g)/d)+4)
	end

	local saturation
	if cmax == 0 then
		saturation = 0
	else
		saturation = d/cmax
	end

	local value = cmax

	return hue, saturation, value
end

local function SetWheelPointer(x, y)
	if not(Healcsi.ColorPicker) then return end

	local pointer = Healcsi.ColorPicker.content.wheelFrame.pointer
	pointer:ClearAllPoints()
	pointer:SetPoint("CENTER", x, y)
end

local function SetResultFrame(r, g, b, a)
	if not(Healcsi.ColorPicker) then return end

	Healcsi.ColorPicker.content.resultFrame.color:SetVertexColor(r, g, b, a or 1)
end

local function SetLightnessSlider(r, g, b)
	if not(Healcsi.ColorPicker) then return end

	local lightnessBg = Healcsi.ColorPicker.content.lightnessFrame.bg

	for i=1, #lightnessBg do
		local k = (-1/200) * i + (201/200)
		lightnessBg[i]:SetVertexColor(k*r, k*g, k*b)
	end
end

local function SetLightnessSliderValue(value)
	local arrows = Healcsi.ColorPicker.content.lightnessFrame.arrows

	arrows:ClearAllPoints()
	arrows:SetPoint("TOPLEFT", -21, 18-value)
end

local function SetHexText(r, g, b)
	Healcsi.ColorPicker.content.hexBox:SetText(RGBToHex(r, g, b))
end

local function CoordinatesToRGB(x, y)
	if not(x) or not(y) then return end

	local saturation = math.min(sqrt((x^2)+(y^2)) / 128, 1)

	local hue = RadToDeg(atan(x / y))

	if (y < 0) then
		hue = 180 + hue
	elseif (x < 0 and y > 0) then
		hue = 360 + hue 
	end

	return HSV2RGB(hue, saturation, 1)
end

local function CoordinatesToHS(x, y)
	if not(x) or not(y) then return end

	local saturation = math.min(sqrt((x^2)+(y^2)) / 128, 1)

	local hue = RadToDeg(atan(x / y))

	if (y < 0) then
		hue = 180 + hue
	elseif (x < 0 and y > 0) then
		hue = 360 + hue 
	end

	return hue, saturation
end

local function RGBToCoordinates(r, g, b)
	local hue, saturation = RGB2HSV(r, g, b)
	local x, y = 0, 0
	x = math.sin(DegToRad(hue)) * saturation * 128
	y = math.cos(DegToRad(hue)) * saturation * 128
	return x, y
end

local function HSToCoordinates(h, s)
	local x, y = 0, 0
	x = math.sin(DegToRad(h)) * s * 128
	y = math.cos(DegToRad(h)) * s * 128
	return x, y
end

--[[-----------------------------------------------------------------------------
Scripts
-------------------------------------------------------------------------------]]

local function CancelButton_OnClick()
	if not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.control) then return end
	--PlaySound("gsTitleOptionExit")

	-- restore init state
	local initColor = Healcsi.ColorPicker.init
	local r, g, b, a = initColor.r, initColor.g, initColor.b, initColor.a

	Healcsi.ColorPicker.control:Fire("OnValueChanged", r, g, b, a)

	Healcsi.ColorPicker:Hide()
	Healcsi.ColorPicker.control = nil
	Healcsi.ColorPicker.handlerName = nil
end

local function OkButton_OnClick()
	if not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.control) then return end

	-- the latest update should have already pushed the last value, just close the window here

	Healcsi.ColorPicker:Hide()
	Healcsi.ColorPicker.control = nil
	Healcsi.ColorPicker.handlerName = nil
end

local function Window_OnMouseDown()
	AceGUI:ClearFocus()
end

local function WindowTitle_OnMouseDown()
	if not(Healcsi.ColorPicker) then return end
	Healcsi.ColorPicker:StartMoving()
	AceGUI:ClearFocus()

	local title = Healcsi.ColorPicker.title
	title.isMoving = true
end

local function WindowTitle_OnMouseUp()
	if not(Healcsi.ColorPicker) then return end
	Healcsi.ColorPicker:StopMovingOrSizing()

	local x, y = RGBToCoordinates(Healcsi.ColorPicker.r, Healcsi.ColorPicker.g, Healcsi.ColorPicker.b)
	SetWheelPointer(x, y)

	local title = Healcsi.ColorPicker.title
	title.isMoving = false
end

local function WindowTitle_OnUpdate(self, elapsed)
	if not(Healcsi.ColorPicker) then return end

	local title = Healcsi.ColorPicker.title
	if title.isMoving then

		title.updateElapsed = (title.updateElapsed or 0) + elapsed
		if title.updateElapsed < 0.1 then
			return
		end
		title.updateElapsed = 0

		local x, y = RGBToCoordinates(Healcsi.ColorPicker.r, Healcsi.ColorPicker.g, Healcsi.ColorPicker.b)
		SetWheelPointer(x, y)
	end
end

local function UpdateHSVColor()
	if not(Healcsi.ColorPicker) then return end

	local h, s, v, a = Healcsi.ColorPicker.h, Healcsi.ColorPicker.s, Healcsi.ColorPicker.v, Healcsi.ColorPicker.a

	if not(v) then
		v = 1
		SetLightnessSliderValue(0)
	end

	if not(h) or not(s) then
		h = 0
		s = 0
	end

	local r, g, b = HSV2RGB(h, s, 1)
	SetLightnessSlider(r, g, b)

	r, g, b       = HSV2RGB(h, s, 1-v)
	local content = Healcsi.ColorPicker.content

	-- update sliders
	content.rSlider:SetValue(round(r*255,0))
	content.gSlider:SetValue(round(g*255,0))
	content.bSlider:SetValue(round(b*255,0))
	content.aSlider:SetValue(round(a*255,0))

	-- update internal storage
	Healcsi.ColorPicker.r = r
	Healcsi.ColorPicker.g = g
	Healcsi.ColorPicker.b = b

	-- update result rectangle
	SetResultFrame(r, g, b, a)

	-- update hex
	SetHexText(r, g, b)

	--Healcsi:Print("U.HSV fires: " ..r.." | "..g.." | "..b.." -- "..a, 1,0.8,0.3)
	Healcsi.ColorPicker.control:Fire("OnValueChanged", r, g, b, a)
end

local function UpdateRGBColor(skipFire)
	if not(Healcsi.ColorPicker) then return end

	local r, g, b, a = Healcsi.ColorPicker.r, Healcsi.ColorPicker.g, Healcsi.ColorPicker.b, Healcsi.ColorPicker.a
	local content = Healcsi.ColorPicker.content

--[[
	if(skipFire) then
		Healcsi:Print("U.RGB INIT sees: " ..r.." | "..g.." | "..b.." -- "..a, 1,0.5,0.5)
	else
		Healcsi:Print("U.RGB sees: " ..r.." | "..g.." | "..b.." -- "..a, 1,0,0)
	end
]]

	-- update sliders
	content.rSlider:SetValue(round(r*255,0))
	content.gSlider:SetValue(round(g*255,0))
	content.bSlider:SetValue(round(b*255,0))
	content.aSlider:SetValue(round(a*255,0))

	-- update result rectangle
	SetResultFrame(r, g, b, a)

	-- update hex
	SetHexText(r, g, b)

	-- update wheel
	local h, s, v = RGB2HSV(r, g, b)
	local x, y = HSToCoordinates(h, s)
	SetWheelPointer(x, y)

	-- update internal storage
	Healcsi.ColorPicker.h = h
	Healcsi.ColorPicker.s = s
	Healcsi.ColorPicker.v = (1-v)

	-- update lightness
	SetLightnessSlider(HSV2RGB(h, s, 1))
	SetLightnessSliderValue((1-v)*200)

	--Healcsi:Print("U.RGB no fire: " ..r.." | "..g.." | "..b.." -- "..a, 0.3,0.8,0.5)

	if skipFire then return end

	--Healcsi:Print("U.RGB fires: " ..r.." | "..g.." | "..b.." -- "..a, 1,0.8,0.3)
	Healcsi.ColorPicker.control:Fire("OnValueChanged", r, g, b, a)
end

local function RedColorChanged(self, event, value)
	if not(Healcsi.ColorPicker) then return end

	Healcsi.ColorPicker.r = value/255
	UpdateRGBColor()
end

local function GreenColorChanged(self, event, value)
	if not(Healcsi.ColorPicker) then return end

	Healcsi.ColorPicker.g = value/255
	UpdateRGBColor()
end

local function BlueColorChanged(self, event, value)
	if not(Healcsi.ColorPicker) then return end

	Healcsi.ColorPicker.b = value/255
	UpdateRGBColor()
end

local function AlphaColorChanged(self, event, value)
	if not(Healcsi.ColorPicker) then return end

	Healcsi.ColorPicker.a = value/255
	UpdateRGBColor()
end

local function IsStringHexadecimal(text)
	for i=1,text:len() do
		local char =text:sub(i,i)
		if(
			char ~= '0' and
			char ~= '1' and
			char ~= '2' and
			char ~= '3' and
			char ~= '4' and
			char ~= '5' and
			char ~= '6' and
			char ~= '7' and
			char ~= '8' and
			char ~= '9' and
			char ~= 'a' and char ~= 'A' and
			char ~= 'b' and char ~= 'B' and
			char ~= 'c' and char ~= 'C' and
			char ~= 'd' and char ~= 'D' and
			char ~= 'e' and char ~= 'E' and
			char ~= 'f' and char ~= 'F'
		)then
			return false
		end
	end
	return true
end

local function HexColorChanged(self, event, value)
	if not(Healcsi.ColorPicker) then return end

	if (string.len(value) == 7 and value:sub(1,1) == '#') then
		value = value:sub(2,7)
	end

	if IsStringHexadecimal(value) then
		local hexValue = nil

		if (string.len(value) == 6) then
			hexValue = value
		elseif (string.len(value) == 3) then
			hexValue = value:sub(1,1)..value:sub(1,1)..value:sub(2,2)..value:sub(2,2)..value:sub(3,3)..value:sub(3,3)
		end

		if hexValue then

			local r, g, b = HexToRGB(hexValue)
			Healcsi.ColorPicker.r = r
			Healcsi.ColorPicker.g = g
			Healcsi.ColorPicker.b = b

			UpdateRGBColor()
		end
	end
end


local function ColorWheel_OnMouseDown()
	if not(Healcsi.ColorPicker) then return end

	local pointer = Healcsi.ColorPicker.content.wheelFrame.pointer

	local x, y = GetWheelClickCoordinates()
	SetWheelPointer(x, y)

	local h, s = CoordinatesToHS(x, y)
	Healcsi.ColorPicker.h = h
	Healcsi.ColorPicker.s = s
	UpdateHSVColor()
		
	pointer.isMoving = true
	pointer.updateElapsed = 0
	pointer:StartMoving()
	
	AceGUI:ClearFocus()
end

local function ColorWheel_OnMouseUp()
	if not(Healcsi.ColorPicker) then return end

	local pointer = Healcsi.ColorPicker.content.wheelFrame.pointer
	
	local x, y = GetWheelClickCoordinates()
	SetWheelPointer(x, y)

	local h, s = CoordinatesToHS(x, y)
	Healcsi.ColorPicker.h = h
	Healcsi.ColorPicker.s = s
	UpdateHSVColor()

	pointer:StopMovingOrSizing()
	pointer.isMoving = false
end

local function ColorWheel_OnUpdate(self, elapsed)
	if not(Healcsi.ColorPicker) then return end

	local pointer = Healcsi.ColorPicker.content.wheelFrame.pointer
	if pointer.isMoving then

		pointer.updateElapsed = (pointer.updateElapsed or 0) + elapsed
		if pointer.updateElapsed < 0.1 then
			return
		end
		pointer.updateElapsed = 0

		local x, y = GetWheelClickCoordinates()
		SetWheelPointer(x, y)

		local h, s = CoordinatesToHS(x, y)
		Healcsi.ColorPicker.h = h
		Healcsi.ColorPicker.s = s
		UpdateHSVColor()
	end
end

local function Lightness_OnMouseDown()
	if not(Healcsi.ColorPicker) then return end
	
	local lightnessFrame = Healcsi.ColorPicker.content.lightnessFrame
	lightnessFrame.isMoving = true

	local v = GetLightnessClickCoordinate()
	SetLightnessSliderValue(v)
	Healcsi.ColorPicker.v = v/200
	UpdateHSVColor()
end

local function Lightness_OnMouseUp()
	if not(Healcsi.ColorPicker) then return end

	local lightnessFrame = Healcsi.ColorPicker.content.lightnessFrame
	lightnessFrame.isMoving = false

	local v = GetLightnessClickCoordinate()
	SetLightnessSliderValue(v)
	Healcsi.ColorPicker.v = v/200
	UpdateHSVColor()
end

local function Lightness_OnUpdate(self, elapsed)
	if not(Healcsi.ColorPicker) then return end

	local lightnessFrame = Healcsi.ColorPicker.content.lightnessFrame
	if lightnessFrame.isMoving then

		lightnessFrame.updateElapsed = (lightnessFrame.updateElapsed or 0) + elapsed
		if lightnessFrame.updateElapsed < 0.1 then
			return
		end
		lightnessFrame.updateElapsed = 0

		local v = GetLightnessClickCoordinate()
		SetLightnessSliderValue(v)
		Healcsi.ColorPicker.v = v/200
		UpdateHSVColor()
	end
end

local function ReRegisterControl(self)
	Healcsi.ColorPicker.control = self


	Healcsi.ColorPicker.r = self.r
	Healcsi.ColorPicker.g = self.g
	Healcsi.ColorPicker.b = self.b
	Healcsi.ColorPicker.a = self.a


	if self.HasAlpha then
		Healcsi.ColorPicker.content.aSlider:SetDisabled(false)
	else
		Healcsi.ColorPicker.content.aSlider:SetValue(255)
		Healcsi.ColorPicker.content.aSlider:SetDisabled(true)
	end

	UpdateRGBColor(true)
end

-- Input scripts

local function OnEnter(hover)
	local frame = hover.parentFrame
	frame.obj:Fire("OnEnter")
	frame.background:SetVertexColor(1, 1, 0)
	frame.background:SetVertexColor(Healcsi:GetColor("BlueHovered", 0.9))
end

local function OnLeave(hover)
	local frame = hover.parentFrame
	frame.obj:Fire("OnLeave")
	frame.background:SetVertexColor(Healcsi:GetColor("DarkBlue", 0.9))
end

local function OnClickPass(hover, ...)
	local frame = hover.parentFrame
	frame:GetScript("OnClick")(frame, ...)
end

local function OnClick(frame)

	if not(Healcsi.ColorPicker) then return end

	local self = frame.obj
	if (not(self.disabled) and self.handlerName) then

		Healcsi.ColorPicker.control     = self
		Healcsi.ColorPicker.handlerName = self.handlerName

		Healcsi.ColorPicker.init = {}
		Healcsi.ColorPicker.init.r = self.r
		Healcsi.ColorPicker.init.g = self.g
		Healcsi.ColorPicker.init.b = self.b
		Healcsi.ColorPicker.init.a = self.a

		Healcsi.ColorPicker.r = self.r
		Healcsi.ColorPicker.g = self.g
		Healcsi.ColorPicker.b = self.b
		Healcsi.ColorPicker.a = self.a

		Healcsi.ColorPicker:Show()

		if self.HasAlpha then
			Healcsi.ColorPicker.content.aSlider:SetDisabled(false)
		else
			Healcsi.ColorPicker.content.aSlider:SetValue(255)
			Healcsi.ColorPicker.content.aSlider:SetDisabled(true)
		end

		UpdateRGBColor(true)
	end
end


--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
	["OnAcquire"] = function(self)
		self.handlerName = nil -- clear handler name so no update can be done at all while things are not initialized

		self:SetFont(Healcsi:GetFont("Friz Quadrata TT"))
		self:SetFontColor(Healcsi:GetColor("Black", 1))

		self:SetShadowColor(Healcsi:GetColor("DarkBlue", 0.5))
		self:SetShadowOffset(1, -1)

		self:SetHeight(24)
		self:SetWidth(200)
		self:SetHasAlpha(false)
		--self:SetRGBAColor(0, 0, 0, 1) -- Screwed up color init
		self:SetDisabled(nil)
		self:SetLabel(nil)
	end,

	-- ["OnRelease"] = nil,

	["SetLabel"] = function(self, text)
		self.text:SetText(text)
	end,

	["SetRGBColor"] = function(self, r, g, b)
		self.r = r
		self.g = g
		self.b = b
		self.a = 1
		self.frame.swatch:SetVertexColor(r, g, b, 1)

		if (not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.handlerName) or not(self.handlerName)) then return end

		if (Healcsi.ColorPicker.handlerName == self.handlerName) then
			Healcsi.ColorPicker.r = r
			Healcsi.ColorPicker.g = g
			Healcsi.ColorPicker.b = b
			Healcsi.ColorPicker.a = a or 1
		end

		UpdateRGBColor(true) -- was missing...
	end,

	["SetRGBAColor"] = function(self, r, g, b, a)
		self.r = r
		self.g = g
		self.b = b
		self.a = a or 1
		self.frame.swatch:SetVertexColor(r, g, b, a)

		if (not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.handlerName) or not(self.handlerName)) then return end

		if (Healcsi.ColorPicker.handlerName == self.handlerName) then
			Healcsi.ColorPicker.r = r
			Healcsi.ColorPicker.g = g
			Healcsi.ColorPicker.b = b
			Healcsi.ColorPicker.a = a or 1

			--Healcsi:Print("SetRGBAColor: " .. self.handlerName .. " --> " ..round(r*255,0).." : "..round(g*255,0).." : "..round(b*255,0).." : "..round(a*255,0), 0.3, 0.6, 0.4)
		end

		UpdateRGBColor(true) -- was missing...
	end,

	["SetRGBAColorObj"] = function(self, color)
		color = color or {}
		self.r = color.r or 0
		self.g = color.g or 0
		self.b = color.b or 0
		self.a = color.a or 1
		self.frame.swatch:SetVertexColor(self.r, self.g, self.b, self.a)

		if (not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.handlerName) or not(self.handlerName)) then return end

		if (Healcsi.ColorPicker.handlerName == self.handlerName) then
			Healcsi.ColorPicker.r = self.r
			Healcsi.ColorPicker.g = self.g
			Healcsi.ColorPicker.b = self.b
			Healcsi.ColorPicker.a = self.a

			--Healcsi:Print("SetRGBAColor: " .. self.handlerName .. " --> " ..round(self.r*255,0).." : "..round(self.g*255,0).." : "..round(self.b*255,0).." : "..round(self.a*255,0), 0.3, 0.6, 0.4)
		end

		UpdateRGBColor(true) -- was missing...
	end,

	["SetHexColor"] = function(self, hex)
		local r, g, b, a = HexToRGBA(hex)
		self.r = r
		self.g = g
		self.b = b
		self.a = a or 1
		self.frame.swatch:SetVertexColor(r, g, b, a)

		if (not(Healcsi.ColorPicker) or not(Healcsi.ColorPicker.handlerName) or not(self.handlerName)) then return end

		if (Healcsi.ColorPicker.handlerName == self.handlerName) then
			Healcsi.ColorPicker.r = r
			Healcsi.ColorPicker.g = g
			Healcsi.ColorPicker.b = b
			Healcsi.ColorPicker.a = a or 1
		end

	end,

	["SetHasAlpha"] = function(self, HasAlpha)
		self.HasAlpha = HasAlpha
	end,

	["SetHandlerName"] = function(self, name)
		self.handlerName = name

		if not(Healcsi.ColorPicker) then return end

		if (Healcsi.ColorPicker.handlerName and Healcsi.ColorPicker.handlerName == self.handlerName) then
			ReRegisterControl(self)-- VERY important to re-assign handler resource!
		end
	end,

	["SetDisabled"] = function(self, disabled)
		self.disabled = disabled
		if self.disabled then
			self.frame:Disable()
			self.text:SetTextColor(0.2, 0.2, 0.4)
		else
			self.frame:Enable()
			local color = self.text.color or {}
			self.text:SetTextColor(color.r or 1, color.g or 1, color.b or 1, color.a or 1 )
			--self.text:SetTextColor(1, 1, 1)
		end
	end,

	["SetFont"] = function(self, font, height, flags)
		self.text:SetFont(Healcsi:GetMedia("font", font) or font, height or 14, flags)
	end,

	["SetFontColor"] = function(self, r, g, b, a)
		self.text:SetVertexColor(r or 0, g or 0, b or 0, a or 1)
		self.text.color = {}
		self.text.color.r = r or 0
		self.text.color.g = g or 0
		self.text.color.b = b or 0
		self.text.color.a = a or 0
	end,

	["SetShadowColor"] = function(self, r, g, b, a)
		self.text:SetShadowColor(r or 0, g or 0, b or 0, a or 1)
	end,

	["SetShadowOffset"] = function(self, x, y)
		self.text:SetShadowOffset(x or 0, y or 0)
	end

}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]

local FrameBackdrop = {
	bgFile = Healcsi:GetMedia("background", "Solid"),
	edgeFile = nil,
	tile = true, tileSize = 32, edgeSize = 32,
	insets = { left = 0, right = 0, top = 0, bottom = 0 }
}

local function Constructor()
	local num = AceGUI:GetNextWidgetNum(Type)

	local frame = CreateFrame("Button", "HealcsiColorPicker" .. num, Healcsi.UIParent)
	frame:Hide()

	frame:SetSize(24,24)
	frame:EnableMouse(true)
	frame:SetScript("OnClick", OnClick)

	local hover = CreateFrame("Button", nil, frame)
	hover.parentFrame = frame
	hover:SetAllPoints()
	hover:SetScript('OnEnter', OnEnter)
	hover:SetScript('OnLeave', OnLeave)
	hover:SetScript('OnClick', OnClickPass)
	frame.hover = hover

	local swatch = frame:CreateTexture(nil, "ARTWORK")
	swatch:SetPoint("CENTER", frame, "LEFT", 12, 0)
	swatch:SetSize(20, 20)
	swatch:SetTexture(Healcsi:GetMedia("background", "Solid"))
	swatch:SetVertexColor(1, 1, 1)
	frame.swatch = swatch

	local background = frame:CreateTexture(nil, 'OVERLAY', nil, 1)
	background:SetPoint('CENTER', swatch)
	background:SetSize(24, 24)
	background:SetTexture([[Interface\Addons\Healcsi\Media\ColorPicker\swatch]])
	background:SetVertexColor(Healcsi:GetColor("DarkBlue", 0.9))
	frame.background = background

	local checkers = frame:CreateTexture(nil, "BACKGROUND")
	checkers:SetPoint("CENTER", swatch)
	checkers:SetSize(20, 20)
	checkers:SetTexture("Tileset\\Generic\\Checkers")
	checkers:SetTexCoord(.25, 0, 0.5, .25)
	checkers:SetDesaturated(true)
	checkers:SetVertexColor(1, 1, 1, 0.75)
	frame.checkers = checkers

	local text = frame:CreateFontString(nil,"OVERLAY")
	text:SetHeight(24)
	text:SetJustifyH("LEFT")
	text:SetJustifyV("MIDDLE")
	text:SetPoint("LEFT", swatch, "RIGHT", 8, 0)
	text:SetPoint("RIGHT")

	if not(Healcsi.ColorPicker) then

		local window = CreateFrame("Frame", nil, Healcsi.UIParent)
		window:Hide()

		window:EnableMouse(true)
		window:SetPoint("CENTER", Healcsi.UIParent, "CENTER")
		window:SetMovable(true)
		window:SetResizable(false)
		window:SetFrameStrata("FULLSCREEN_DIALOG")
		window:SetBackdrop(FrameBackdrop)
		window:SetBackdropColor(Healcsi:GetColor("BlueBackground", 0.9))
		window:SetMinResize(400, 200)
		window:SetSize(640, 340)
		window:SetToplevel(true)
		window:SetScript("OnMouseDown", Window_OnMouseDown)


		Healcsi:SetFrameShadow(window, "Shadow-3")
		Healcsi:SetFrameShadowColor(window, Healcsi:GetColor("DarkBlue", 1))

		local cancelbutton = AceGUI:Create("Healcsi-Button")
		cancelbutton:SetCallback("OnClick", CancelButton_OnClick)
		cancelbutton:SetHeight(20)
		cancelbutton:SetWidth(100)
		cancelbutton.frame:SetParent(window)
		cancelbutton.frame:SetPoint("BOTTOMRIGHT", window, "BOTTOMRIGHT", -27, 18) -- -27px left, 18px up
		cancelbutton:SetText(L["Cancel"])
		cancelbutton.frame:Show()

		local okbutton = AceGUI:Create("Healcsi-Button")
		okbutton:SetCallback("OnClick", OkButton_OnClick)
		okbutton:SetHeight(20)
		okbutton:SetWidth(100)
		okbutton.frame:SetParent(window)
		okbutton.frame:SetPoint("BOTTOMRIGHT", window, "BOTTOMRIGHT", -150, 18) -- -150px left, 18px up
		okbutton:SetText(L["Okay"])
		okbutton.frame:Show()

		local titlebg = window:CreateTexture(nil, "OVERLAY")
		titlebg:SetTexture(Healcsi:GetMedia("background", "Solid"))
		titlebg:SetVertexColor(Healcsi:GetColor("DarkBlue", 1))
		titlebg:SetPoint("TOP", 0, 8)
		titlebg:SetWidth(200)
		titlebg:SetHeight(26)

		local title = CreateFrame("Frame", nil, window)
		title:EnableMouse(true)
		title:SetScript("OnMouseDown", WindowTitle_OnMouseDown)
		title:SetScript("OnMouseUp", WindowTitle_OnMouseUp)
		title:SetScript("OnUpdate", WindowTitle_OnUpdate)
		title:SetAllPoints(titlebg)
		title.isMoving = false

		local titletext = title:CreateFontString(nil, "OVERLAY")
		titletext:SetPoint("TOP", titlebg, "TOP", 0, -6)
		titletext:SetFont(Healcsi:GetMedia("font", "Friz Quadrata TT") , 15)
		titletext:SetVertexColor(Healcsi:GetColor("BlueSelected", 1))
		titletext:SetShadowColor(Healcsi:GetColor("DarkBlue", 0.5))
		titletext:SetShadowOffset(1, -1)
		titletext:SetText("Color Picker")

		--Container Support
		local content = CreateFrame("Frame", nil, window)
		content:SetPoint("TOPLEFT", 17, -27) -- 17px right , -27px down
		content:SetPoint("BOTTOMRIGHT", -17, 20) -- 17px left, 40px up

		--content.bgTex = Healcsi:SetMiddleBackground(content, "ARTWORK", "ShamanBackground", 1024, 793, -15, 1024)
		--content.bgTex:SetVertexColor(1, 1, 1, 0.3)

		window.cancelbutton  = cancelbutton
		window.okbutton      = okbutton
		window.title         = title
		window.title.text    = titletext
		window.content       = content

		local iconTexture = content:CreateTexture(nil, "OVERLAY")
		iconTexture:SetPoint("TOPLEFT" , 3 , 4)
		iconTexture:SetSize(32, 32)
		iconTexture:SetTexture("Interface\\Addons\\Healcsi\\Media\\ColorPicker\\Icon")
		content.icon = iconTexture

		-- COLOR WHEEL:
		local wheelFrame = CreateFrame("Button", nil, window.content)
		wheelFrame:SetSize(256,256)
		wheelFrame:SetPoint("BOTTOMLEFT", 2, 15) -- 2px right, 10px up
		wheelFrame:SetMovable(true)
		wheelFrame:SetScript("OnMouseDown", ColorWheel_OnMouseDown)
		wheelFrame:SetScript("OnMouseUp", ColorWheel_OnMouseUp)
		wheelFrame:SetScript("OnUpdate", ColorWheel_OnUpdate)

		local wheel = wheelFrame:CreateTexture(nil, "OVERLAY")
		wheel:SetAllPoints(wheelFrame)
		wheel:SetTexture("Interface\\Addons\\Healcsi\\Media\\ColorPicker\\ColorWheel")
		wheelFrame.wheel = wheel

		local pointer = CreateFrame("Button", nil, wheelFrame)
		pointer:SetPoint("CENTER")
		pointer:SetSize(16,16)
		pointer:SetMovable(true)
		pointer:SetScript("OnMouseDown", ColorWheel_OnMouseDown)
		pointer:SetScript("OnMouseUp", ColorWheel_OnMouseUp)
		pointer:SetScript("OnUpdate", ColorWheel_OnUpdate)
		pointer.isMoving = false
		wheelFrame.pointer = pointer

		local pointerTexture = pointer:CreateTexture(nil, "BACKGROUND")
		pointerTexture:SetAllPoints(pointer)
		pointerTexture:SetTexture("Interface\\Addons\\Healcsi\\Media\\ColorPicker\\Pointer")
		wheelFrame.pointer.texture = pointerTexture

		window.content.wheelFrame = wheelFrame

		local resultFrame = CreateFrame("Button", nil, window.content)
		resultFrame:SetPoint("TOPLEFT", wheelFrame, "TOPRIGHT", 30, 10) -- 30px right, 10px down
		resultFrame:SetSize(54,36)

		local resultColor = resultFrame:CreateTexture(nil, "OVERLAY")
		resultColor:SetAllPoints(resultFrame)
		resultColor:SetTexture("Interface\\Buttons\\WHITE8X8")
		resultColor:SetVertexColor(1, 1, 1)
		resultFrame.color = resultColor

		local resultCheckers = resultFrame:CreateTexture(nil, "BACKGROUND")
		--resultCheckers:SetPoint("CENTER", resultColor)
		resultCheckers:SetAllPoints(resultColor)
		resultCheckers:SetTexture("Tileset\\Generic\\Checkers")
		resultCheckers:SetTexCoord(.25, 0.01, 0.5, .25)
		resultCheckers:SetDesaturated(true)
		resultCheckers:SetVertexColor(1, 1, 1, 0.75)
		resultFrame.checkers = resultCheckers

		window.content.resultFrame = resultFrame

		local rSlider = AceGUI:Create("Healcsi-Slider")
		rSlider:SetCallback("OnValueChanged", RedColorChanged)
		rSlider:SetSliderValues(0, 255, 1)
		rSlider:SetLabel("Red")
		rSlider:SetValue(0)
		rSlider:SetWidth(200)
		rSlider:SetHeight(44)
		rSlider.frame:SetParent(window.content)
		rSlider.frame:SetPoint("TOPRIGHT", window.content, "TOPRIGHT", -17, -10)
		window.content.rSlider = rSlider
		rSlider.frame:Show()

		local gSlider = AceGUI:Create("Healcsi-Slider")
		gSlider:SetCallback("OnValueChanged", GreenColorChanged)
		gSlider:SetSliderValues(0, 255, 1)
		gSlider:SetLabel("Green")
		gSlider:SetValue(0)
		gSlider:SetWidth(200)
		gSlider:SetHeight(44)
		gSlider.frame:SetParent(window.content)
		gSlider.frame:SetPoint("TOPRIGHT", window.content, "TOPRIGHT", -17, -60)
		window.content.gSlider = gSlider
		gSlider.frame:Show()

		local bSlider = AceGUI:Create("Healcsi-Slider")
		bSlider:SetCallback("OnValueChanged", BlueColorChanged)
		bSlider:SetSliderValues(0, 255, 1)
		bSlider:SetLabel("Blue")
		bSlider:SetValue(0)
		bSlider:SetWidth(200)
		bSlider:SetHeight(44)
		bSlider.frame:SetParent(window.content)
		bSlider.frame:SetPoint("TOPRIGHT", window.content, "TOPRIGHT", -17, -110)
		window.content.bSlider = bSlider
		bSlider.frame:Show()

		local aSlider = AceGUI:Create("Healcsi-Slider")
		aSlider:SetCallback("OnValueChanged", AlphaColorChanged)
		aSlider:SetSliderValues(0, 255, 1)
		aSlider:SetLabel("Alpha")
		aSlider:SetValue(0)
		aSlider:SetWidth(200)
		aSlider:SetHeight(44)
		aSlider.frame:SetParent(window.content)
		aSlider.frame:SetPoint("TOPRIGHT", window.content, "TOPRIGHT", -17, -160)
		window.content.aSlider = aSlider
		aSlider.frame:Show()

		local hexBox = AceGUI:Create("Healcsi-EditBox")
		hexBox:SetLabel("Hex color code:")
		hexBox:SetCallback("OnEnterPressed", HexColorChanged)
		hexBox:SetWidth(200)
		hexBox:SetHeight(44)
		hexBox.frame:SetParent(window.content)
		hexBox.frame:SetPoint("TOPRIGHT", window.content, "TOPRIGHT", -17, -210)
		hexBox.label:SetFontObject("GameFontNormal") 
		hexBox.frame:Show()
		window.content.hexBox = hexBox

		-- lightness

		local lightnessFrame = CreateFrame("Button", nil, window.content)
		lightnessFrame:SetPoint("TOPLEFT", wheelFrame, "TOPRIGHT", 34, -50) -- 34px right, 50px down
		lightnessFrame:SetSize(46,200)
		lightnessFrame:SetScript("OnMouseDown", Lightness_OnMouseDown)
		lightnessFrame:SetScript("OnMouseUp", Lightness_OnMouseUp)
		lightnessFrame:SetScript("OnUpdate", Lightness_OnUpdate)
		lightnessFrame.isMoving = false
		
		local lightnessArrows = lightnessFrame:CreateTexture(nil, "OVERLAY")
		lightnessArrows:SetPoint("TOPLEFT", -21, 18)
		lightnessArrows:SetSize(128, 64)
		lightnessArrows:SetTexture("Interface\\Addons\\Healcsi\\Media\\ColorPicker\\Arrows")
		lightnessFrame.arrows = lightnessArrows

		local lightnessBg = {}
		for i=1, 200 do
			local newLightnessFrame = lightnessFrame:CreateTexture(nil, "OVERLAY")
			newLightnessFrame.bgId = i
			newLightnessFrame:SetSize(46,1)
			newLightnessFrame:SetPoint("TOP", lightnessFrame, "TOP", 0, -i)
			newLightnessFrame:SetTexture("Interface\\Buttons\\WHITE8X8")

			local k = (-1/200) * i + (201/200)
			newLightnessFrame:SetVertexColor(k, k, k)

			table.insert(lightnessBg, newLightnessFrame) -- insert to end of table
		end
		lightnessFrame.bg = lightnessBg
		window.content.lightnessFrame = lightnessFrame

		Healcsi.ColorPicker = window
	end

	local widget = {
		swatch   = swatch,
		text     = text,
		frame    = frame,
		type     = Type
	}
	for method, func in pairs(methods) do
		widget[method] = func
	end

	return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
