
local Healcsi = Healcsi
local Gui     = Healcsi.Gui
local L       = Healcsi.Locale
local Editor  = Healcsi.Editor

local Profiles = {}
local contentHolder = nil

local defaultProfiles
local tmpprofiles = {}

--[[ fill defaultProfiles with some generic values ]]
function GenerateDefaultProfiles(db)
	defaultProfiles = {
		["Default"] = L["DefaultProfile"],
		[db.keys.char] = db.keys.char,
		[db.keys.realm] = db.keys.realm,
		[db.keys.class] = UnitClass("player")
	}
end

-- Get a list of available profiles for the specified database.
-- You can specify which profiles to include/exclude in the list using the two boolean parameters listed below.
-- @param db The db object to retrieve the profiles from
-- @param common If true, getProfileList will add the default profiles to the return list, even if they have not been created yet
-- @param nocurrent If true, then getProfileList will not display the current profile in the list
-- @return Hashtable of all profiles with the internal name as keys and the display name as value.
function GetProfileList(db, common, nocurrent)
	local profiles = {}
	
	-- copy existing profiles into the table
	local currentProfile = db:GetCurrentProfile()
	for i,v in pairs(db:GetProfiles(tmpprofiles)) do 
		if not (nocurrent and v == currentProfile) then 
			profiles[v] = v 
		end 
	end

	if not defaultProfiles then
		GenerateDefaultProfiles(db)
	end
	
	-- add our default profiles to choose from ( or rename existing profiles)
	for k,v in pairs(defaultProfiles) do
		if (common or profiles[k]) and not (nocurrent and k == currentProfile) then
			profiles[k] = v
		end
	end
	
	return profiles
end

--[[ 
	List all active profiles
	you can control the output with the .arg variable
	currently four modes are supported
	
	(empty) - return all available profiles
	"nocurrent" - returns all available profiles except the currently active profile
	"common" - returns all avaialble profiles + some commonly used profiles ("char - realm", "realm", "class", "Default")
	"both" - common except the active profile
]]
function ListProfiles(info)
	if (not Healcsi.db) then return {} end
	local profiles
	if info == "common" then
		profiles = GetProfileList(Healcsi.db, true, nil)
	elseif info == "nocurrent" then
		profiles = GetProfileList(Healcsi.db, nil, true)
	elseif info == "both" then
		profiles = GetProfileList(Healcsi.db, true, true)
	else
		profiles = GetProfileList(Healcsi.db)
	end
	
	return profiles
end

function Profiles:RefreshProfilePage()
	self.data.current:SetText("  " .. L["CurrentProfile"] .. "|cFFFF4500" .. Healcsi.db:GetCurrentProfile() .. "|r")

	self.data.choose:SetList(ListProfiles("common"))
	self.data.choose:SetValue(Healcsi.db:GetCurrentProfile())

	self.data.copy:SetList(ListProfiles("nocurrent"))
	self.data.copy:SetValue("")

	self.data.delete:SetList(ListProfiles("nocurrent"))
	self.data.delete:SetValue("")
end

function Profiles:UpdateDualSpec()
	self.data.checkBox:SetValue(Healcsi.db:IsDualSpecEnabled())

	local _, _, classId = UnitClass("player")

	for i = 1, self.data.numSpecs do
		local currSpec = self.data.dropdowns[i]

		if (Healcsi.db:IsDualSpecEnabled()) then
			currSpec:SetDisabled(false)
			if (currSpec:GetValue() == nil or currSpec:GetValue() == "") then
				currSpec:SetValue(Healcsi.db:GetCurrentProfile())
			end
		else
			currSpec:SetDisabled(true)
			currSpec:SetValue("")
		end

		local name = ""
		local _, specName = GetSpecializationInfoForClassID(classId, i)
		if (Healcsi.DualSpec.currentSpec == i) then
			name = specName .. " (Current)"
		else
			name = specName
		end

		currSpec:SetLabel(name)
	end

	self:RefreshProfilePage()
end

local function DualSpecEventHandler(self, event)
	if event == "PLAYER_LOGIN" then
		self:UnregisterEvent(event)
		self:RegisterUnitEvent("PLAYER_SPECIALIZATION_CHANGED", "player")
	end

	Profiles:UpdateDualSpec()
end

local eventFrame = CreateFrame("Frame")
eventFrame:SetScript("OnEvent", DualSpecEventHandler)

if IsLoggedIn() then
	DualSpecEventHandler(eventFrame, "PLAYER_LOGIN")
else
	eventFrame:RegisterEvent("PLAYER_LOGIN")
end	

function Profiles:CreateSpecEditorView(frame)
	
	Editor:CreateHorizLine(frame)

	local holder = Editor:CreateGroup(frame, true, nil, {10, 15, 10, 15}, {0, 5, 0, 5})

	local specCheckBox = Gui:Create("Healcsi-CheckBox")
	specCheckBox:SetFullWidth(true)
	specCheckBox:SetLabel(L["EnableSpecProfiles"])
	specCheckBox:SetDescription(L["SpecProfilesDesc"])

	specCheckBox:SetLabelFontColor(Healcsi:GetColor("BlueHovered", 1))
	specCheckBox:SetLabelFont(Healcsi:GetFont("Friz Quadrata TT"), 16)
	specCheckBox:SetLabelShadowColor(Healcsi:GetColor("DarkBlue", 0.3))

	specCheckBox:SetValue(Healcsi.db:IsDualSpecEnabled())
	specCheckBox:SetCallback("OnValueChanged", function(self, event, value)
		Healcsi.db:SetDualSpecEnabled(value)
		Profiles:UpdateDualSpec()
	end)

	holder:AddChild(specCheckBox)

	self.data.checkBox = specCheckBox

	local group = Editor:CreateInlineGroup(holder, true, nil, {10, 10, 10, 10}, {0, 5, 0, 5})

	local _, _, classId = UnitClass("player")
	self.data.numSpecs = GetNumSpecializationsForClassID(classId)
	self.data.dropdowns = {}

	for i = 1, self.data.numSpecs do

		local name = ""
		local _, specName = GetSpecializationInfoForClassID(classId, i)
		if (Healcsi.DualSpec.currentSpec == i) then
			name = specName .. " (Current)"
		else
			name = specName
		end

		local currSpec = Gui:Create("Healcsi-Dropdown")
		currSpec:SetList(ListProfiles("common"))
		currSpec:SetValue(Healcsi.db:GetDualSpecProfile(i))
		currSpec:SetCallback("OnValueChanged", function(self, event, value)
			Healcsi.db:SetDualSpecProfile(value, i)
		end)
		currSpec:SetLabel(name)

		if (Healcsi.db:IsDualSpecEnabled()) then
			currSpec:SetDisabled(false)
		else
			currSpec:SetDisabled(true)
			currSpec:SetValue("")
		end

		self.data.dropdowns[i] = currSpec
		group:AddChild(currSpec)

	end

end

function Profiles:CreateEditorView(frame)

	self.data = {}

	local holder = Editor:CreateGroup(frame, true, nil, {10, 0, 10, 0}, {0, 5, 0, 5}, "BlueSelected", 0.3)

	local line1 = Editor:CreateGroup(holder, true, nil, {10, 10, 10, 10}, {0, 5, 5, 5})
	Editor:CreateLabel(line1, true, nil, L["ProfileDesc"], "Black", 1)
	Editor:CreateLabel(line1, true, nil, L["ProfileResetDesc"], "Black", 1)

	local reset = Editor:CreateInlineGroup(line1, true, nil, {2, 5, 2, 5}, {0, 0, 0, 0})

	local resetBtn = Editor:CreateButton(reset, L["ResetProfile"])
	resetBtn:SetCallback("OnClick", function() Healcsi.db:ResetProfile() end)

	self.data.current = Editor:CreateLabel(reset, false, 300, "  " .. L["CurrentProfile"] .. "|cFFFF4500" .. Healcsi.db:GetCurrentProfile() .. "|r", "Black", 1)

	Editor:CreateHorizLine(holder)

	local line2 = Editor:CreateGroup(holder, true, nil, {10, 10, 10, 10}, {0, 5, 0, 5})
	Editor:CreateLabel(line2, true, nil, L["ProfileChoose"], "Black", 1)

	local choose = Editor:CreateInlineGroup(line2, true, nil, {2, 5, 2, 5}, {0, 0, 0, 0})

	local editBox = Editor:CreateEditBox(choose, L["NewProfile"])
	editBox:SetCallback("OnEnterPressed", function(self, event, value)
		Healcsi.db:SetProfile(value)
		Profiles:RefreshProfilePage()
		self:SetText("")
	end)

	local chooseDropDown = Gui:Create("Healcsi-Dropdown")
	chooseDropDown:SetList(ListProfiles("common"))
	chooseDropDown:SetValue(Healcsi.db:GetCurrentProfile())
	chooseDropDown:SetCallback("OnValueChanged", function(self, event, value)
		Healcsi.db:SetProfile(value)
		Profiles:RefreshProfilePage()
	end)
	chooseDropDown:SetLabel(L["ExitingProfiles"])
	choose:AddChild(chooseDropDown)

	self.data.choose = chooseDropDown
	
	if (Healcsi.DualSpec) then
		Profiles:CreateSpecEditorView(holder)
	end

	Editor:CreateHorizLine(holder)

	local line3 = Editor:CreateGroup(holder, true, nil, {10, 10, 10, 10}, {0, 5, 0, 5})
	Editor:CreateLabel(line3, true, nil, L["ProfileCopyDesc"], "Black", 1)

	local copy = Editor:CreateInlineGroup(line3, true, nil, {2, 5, 2, 5}, {0, 0, 0, 0})

	local copyDropDown = Gui:Create("Healcsi-Dropdown")
	copyDropDown:SetList(ListProfiles("nocurrent"))
	copyDropDown:SetLabel(L["CopyProfile"])
	copy:AddChild(copyDropDown)

	copyDropDown:SetCallback("OnValueChanged", function(self, event, value)
		Healcsi.db:CopyProfile(value)
		Profiles:RefreshProfilePage()
	end)

	self.data.copy = copyDropDown

	Editor:CreateHorizLine(holder)

	local line4 = Editor:CreateGroup(holder, true, nil, {10, 10, 10, 10}, {0, 5, 0, 5})
	Editor:CreateLabel(line4, true, nil, L["ProfileDeleteDesc"], "Black", 1)

	local delete = Editor:CreateInlineGroup(line4, true, nil, {2, 5, 2, 5}, {0, 0, 0, 0})

	local delDropDown = Gui:Create("Healcsi-Dropdown")
	delDropDown:SetList(ListProfiles("nocurrent"))
	delDropDown:SetLabel(L["DeleteProfile"])
	delete:AddChild(delDropDown)

	delDropDown:SetCallback("OnValueChanged", function(self, event, value)
		Healcsi.db:DeleteProfile(value)
		Profiles:RefreshProfilePage()
	end)

	self.data.delete = delDropDown

end

function Healcsi:CreateProfileEditor()

	local scrollFrame = Gui:Create("Healcsi-ScrollFrame")
	scrollFrame:SetLayout("Flow")
	scrollFrame:SetFullWidth(true)
	scrollFrame:SetFullHeight(true)

	local parentFrame = Gui:Create("SimpleGroup")
	parentFrame:SetLayout("Flow")
	parentFrame:SetFullWidth(true)
	parentFrame:SetFullHeight(true)

	-- Put page content here

	Profiles:CreateEditorView(parentFrame)

	-- End of page content

	contentHolder = parentFrame

	scrollFrame:AddChild(parentFrame)
	return scrollFrame
end